/**************************
 *         Author         *
 *     Vampyr Yannik      *
 *                        *
 *        4.4.2021        *
 **************************/

#include <mcp_can.h>
#include <SPI.h>

#define lo8(x) ((int)(x)&0xff)
#define hi8(x) ((int)(x)>>8)

int park_brake = A0;

String inputString = "";
boolean stringComplete = false;

//Cluster Variables
boolean signal_check_engine = false;
int temp_signal_check_engine = 0;
boolean diesel_glowplugs = false;
int temp_diesel_glowplugs = 0;
int pack_counter = 0;
boolean backlight = false;
int engine_modules = 0;
int oilpswitch = A1;
boolean oil_pressure_simulation = true;
boolean seat_belt = false;
int temp_seat_belt = 0;
boolean add_distance = false;
int drive_mode = 0;
int distance_multiplier = 2;
int distance_counter = 0;
int turn_signals = 3;
boolean turn_signals_blinking = true;
int turn_signals_counter = 0;
int temp_turn_signals = 0;
int temp_alternator_warning = 0;
boolean alternator_warning = false;
int temp_trunklid = 0;
boolean trunklid = false;
int temp_check = 0;
boolean check = false;
boolean keybattery_warning = false;
int temp_keybattery_warning = 0;
boolean door_open = false;
int lightmode = 0;
boolean clutch_control = false;
int temp_clutch_control = 0;
int other_modules = 0;
boolean airbag = false;
int temp_airbag = 0;
boolean abs_enabled = false;
int temp_abs_enabled = 0;




//Variables for Speed and RPM
int speed = 0;
int speedTmp = 0;
byte speedL = 0;
byte speedH = 0;

int rpm = 0;
short tempRPM = 0;
byte rpmL = 0;
byte rpmH = 0;

const int SPI_CS_PIN = 10;
MCP_CAN CAN(SPI_CS_PIN);

void setup() 
{
  pinMode(park_brake, OUTPUT);
  pinMode(oilpswitch, OUTPUT);

  digitalWrite(park_brake, HIGH);

  Serial.begin(115200);
  while (!Serial) continue;

  //CAN BUS Initialization
START_INIT:

  if (CAN_OK == CAN.begin(CAN_500KBPS, MCP_16MHz))
  {
    Serial.println("CAN BUS Shield init ok!");
  }
  else
  {
    Serial.println("CAN BUS Shield init fail");
    Serial.println("Init CAN BUS Shield again");
    delay(100);
    goto START_INIT;
  }

  //###############################################################

  //Main Setup (Operations)

  //Set constant values to show on cluster
  speed = 0;      //Set the speed in km/h
  rpm = 800;      //set the rev counter
  Fuel(100);      //set the fuel gauge in percent

  //set cluster Functions
  backlight = false;             //Turn on the backlight
  diesel_glowplugs = false;      //Turn on the Glowplug light
  signal_check_engine = false;   //Simulate a error code in the engine module
  turn_signals = 0;             //1=Left, 2=Right, 3=Both
  turn_signals_blinking = true; //Make the turn signals blink or just shine
  airbag = true;                //Airbag Lamp
  abs_enabled = true;           //ABS Module on or off
  alternator_warning = false;   //Generator/Alternator not working
  
  //###############################################################

}

void CanSend(short address, byte a, byte b, byte c, byte d, byte e, byte f, byte g, byte h)
{
  unsigned char DataToSend[8] = {a, b, c, d, e, f, g, h};
  CAN.sendMsgBuf(address, 0, 8, DataToSend);
}

void Fuel(int amount)
{
  pinMode(A4, INPUT);
  pinMode(A5, INPUT);
  pinMode(3, INPUT);
  pinMode(4, INPUT);
  pinMode(5, INPUT);
  pinMode(6, INPUT);
  pinMode(7, INPUT);
  pinMode(8, INPUT);
  pinMode(9, INPUT);

  if (amount >= 87)
  {
    pinMode(A4, OUTPUT);
    digitalWrite(A4, LOW);
    return;
  }
  if (amount >= 75)
  {
    pinMode(A5, OUTPUT);
    digitalWrite(A5, LOW);
    return;
  }
  if (amount >= 62)
  {
    pinMode(3, OUTPUT);
    digitalWrite(3, LOW);
    return;
  }
  if (amount >= 50)
  {
    pinMode(4, OUTPUT);
    digitalWrite(4, LOW);
    return;
  }
  if (amount >= 37)
  {
    pinMode(5, OUTPUT);
    digitalWrite(5, LOW);
    return;
  }
  if (amount >= 25)
  {
    pinMode(6, OUTPUT);
    digitalWrite(6, LOW);
    return;
  }
  if (amount >= 18)
  {
    pinMode(7, OUTPUT);
    digitalWrite(7, LOW);
    return;
  }
  if (amount >= 12)
  {
    pinMode(8, OUTPUT);
    digitalWrite(8, LOW);
    return;
  }
  if (amount >= 0)
  {
    pinMode(9, OUTPUT);
    digitalWrite(9, LOW);
    return;
  }
}

void loop() 
{
  if (stringComplete) {
    Serial.print(inputString.length());
    Serial.print(": ");
    Serial.println(inputString);

    if(inputString.length() >= 39 && inputString.length() <= 46) {
      if(getValue(inputString, ',', 19) != "") {
        speed = getValue(inputString, ',', 0).toInt();
        rpm = getValue(inputString, ',', 1).toInt();
        Fuel(getValue(inputString, ',', 2).toInt());
        
        backlight = getValue(inputString, ',', 3).toInt();

        if(getValue(inputString, ',', 4).toInt() == 1 && getValue(inputString, ',', 5).toInt() == 1) {
          turn_signals = 3;
        } else if (getValue(inputString, ',', 4).toInt()) {
          turn_signals = 1;
        } else if (getValue(inputString, ',', 5).toInt()) {
          turn_signals = 2;
        } else {
          turn_signals = 0;
        }
  

        oil_pressure_simulation = getValue(inputString, ',', 16).toInt();
        Serial.print(speed);
        Serial.print(" - ");
        Serial.print(rpm);
        Serial.print(" - ");
        Serial.println(turn_signals);
      } else {
        Serial.println("INVALID STRING!");
      }
    } else {
        Serial.println("INVALID STRING!");
    }
    
    inputString = "";
    stringComplete = false;
  }
  
  //Conversion Speed
  speedTmp = speed / 0.0075; //KMH=1.12 MPH=0.62

  //Conversion Low and High Bytes
  speedL = lo8(speedTmp);
  speedH = hi8(speedTmp);

  tempRPM = rpm * 4;
  rpmL = lo8(tempRPM);
  rpmH = hi8(tempRPM);  


  if ( oil_pressure_simulation == true)
  {
    //Set Oil Pressure Switch
    if (rpm > 1500) {
      digitalWrite(oilpswitch, LOW);
    }
    else {
      digitalWrite(oilpswitch, HIGH);
    }
  }

  /*Send CAN BUS Informations*/

  //Engine running and ESP on
  CanSend(0xDA0, 0x00, 0x00, speedL, speedH, 0x00, 0x00, 0x00, 0x00);

    //Cruise Control
    //CanSend(0x289, 0x00, B00000101, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00);
    
   //Turning Light blinking
  if (turn_signals_blinking == true)
  {
    if (turn_signals_counter == 15)
    {
      temp_turn_signals = turn_signals;
      turn_signals_counter = turn_signals_counter + 1;
    }
    if (turn_signals_counter == 30)
    {
      temp_turn_signals = 0;
      turn_signals_counter = 0;
    }
    else
    {
      turn_signals_counter = turn_signals_counter + 1;
    }
  }
  else
  {
    temp_turn_signals = turn_signals;
  }

  //Check main settings

  //Engine Modules
  if (diesel_glowplugs == true) temp_diesel_glowplugs = B00000100;
  else temp_diesel_glowplugs = B00000000;
  if (signal_check_engine == true) temp_signal_check_engine = B00001000;
  else temp_signal_check_engine = B00000000;
  engine_modules = temp_diesel_glowplugs + temp_signal_check_engine;

  //Other Modules
  if (airbag == true) CanSend(0x050, 0x00, 0x80, temp_seat_belt, 0x00, 0x00, 0x00, 0x00, 0x00);
  else temp_airbag = B00000000;
  if (abs_enabled == true) CanSend(0x1A0, 0x18, speedL, speedH, 0x00, 0xfe, 0xfe, 0x00, 0xff);
  else temp_abs_enabled = B00000000;
  other_modules = temp_airbag + temp_abs_enabled;

    //Battery Warning
  if (alternator_warning == true) temp_alternator_warning = B10000000;
  else temp_alternator_warning = B00000000;

  
  //Send CAN data every 200ms
  pack_counter++;
  if(pack_counter == 20)
  {

    //Turning Lights 2
    CanSend(0x470, temp_alternator_warning + temp_turn_signals, temp_trunklid + door_open, backlight, 0x00, temp_check + temp_clutch_control, temp_keybattery_warning, 0x00, lightmode);

    //Diesel Engine
    CanSend(0x480, 0x00, engine_modules, 0x00, 0x00, 0x00, temp_diesel_glowplugs, 0x00, 0x00);

    pack_counter = 0;
  }

  //Motorspeed
  CanSend(0x320, 0x00, (speedL * 100), (speedH * 100), 0x00, 0x00, 0x00, 0x00, 0x00);

  //RPM
  CanSend(0x280, 0x49, 0x0E, rpmL, rpmH, 0x0E, 0x00, 0x1B, 0x0E);

  //Speed
  CanSend(0x5A0, 0xFF, speedL, speedH, drive_mode, 0x00, lo8(distance_counter), hi8(distance_counter), 0xad);


  //Wait a time
  delay(5);
}

void serialEvent() {
  while (Serial.available()) {
    char inChar = (char)Serial.read();
    if (inChar == ';') {
      stringComplete = true;
    } else {
      inputString += inChar;
    }
  }
}

String getValue(String data, char separator, int index)
{
  int found = 0;
  int strIndex[] = { 0, -1 };
  int maxIndex = data.length() - 1;
  
  for (int i = 0; i <= maxIndex && found <= index; i++) {
    if (data.charAt(i) == separator || i == maxIndex) {
      found++;
      strIndex[0] = strIndex[1] + 1;
      strIndex[1] = (i == maxIndex) ? i+1 : i;
    }
  }
  return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}
